-- scope stuff
if not classicHeals then
	classicHeals = {};
end
local ch = classicHeals;

-- variables
local eventFrame = CreateFrame("Frame");
eventFrame:RegisterEvent("ADDON_LOADED");
eventFrame:RegisterEvent("PLAYER_ENTERING_WORLD");
eventFrame:RegisterEvent("PLAYER_TARGET_CHANGED");
eventFrame:RegisterEvent("GROUP_ROSTER_UPDATE");
eventFrame:RegisterEvent("RAID_ROSTER_UPDATE");
eventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");

local charName = UnitName("player");
local initFinished = false;

local inRaid = IsInRaid();
local partyMembers = nil;
local raidMembers = nil;

local targetFrame;
local partyFrames = {};
local raidFrames_Compact = {};
local raidFrames_Pullout = {};

local FRAMETYPE_TARGET = 0;
local FRAMETYPE_PARTY = 1;
local FRAMETYPE_RAID = 2;

local OPT_FONTSIZE = "OPT_FONTSIZE";
local OPT_SHOWCRF = "OPT_SHOWCRF";
local OPT_SHOWOVERHEAL = "OPT_SHOWOVERHEAL";

-- command handling
SLASH_CLASSICHEALS1 = "/classicheals";
local CMD_CRF = "crf";
local CMD_FONTSIZE = "fontsize";
local CMD_OVERHEAL = "overheal";

SlashCmdList["CLASSICHEALS"] = function(msg)
	if msg then
		if msg == CMD_CRF then
			CLASSICHEALS_OPTIONS[OPT_SHOWCRF] = not CLASSICHEALS_OPTIONS[OPT_SHOWCRF];
			ch:SetCrfVisible(CLASSICHEALS_OPTIONS[OPT_SHOWCRF]);
			ch:Println(ch:GetText(ch.LKEY_COMMANDS_CRF_SET) .. ch:BoolToYesNo(CLASSICHEALS_OPTIONS[OPT_SHOWCRF]));
			return;
		elseif msg == CMD_OVERHEAL then
			CLASSICHEALS_OPTIONS[OPT_SHOWOVERHEAL] = not CLASSICHEALS_OPTIONS[OPT_SHOWOVERHEAL];
			ch:Println(ch:GetText(ch.LKEY_COMMANDS_OVERHEAL_SET) .. ch:BoolToYesNo(CLASSICHEALS_OPTIONS[OPT_SHOWOVERHEAL]));
			return;
		elseif ch:StartsWith(msg, CMD_FONTSIZE) then
			msg = msg:gsub(CMD_FONTSIZE .. " ", "");
			local fontsize = tonumber(msg);
			if fontsize and fontsize >= 10 and fontsize <= 20 then
				CLASSICHEALS_OPTIONS[OPT_FONTSIZE] = fontsize;
				ch:Println(ch:GetText(ch.LKEY_COMMANDS_FONTSIZE_SET) .. tostring(CLASSICHEALS_OPTIONS[OPT_FONTSIZE]));
				return;
			end
		end
		
	end

	ch:Println(ch:GetText(ch.LKEY_COMMANDS));
	ch:Println(ch:GetText(ch.LKEY_COMMANDS_CRF));
	ch:Println(ch:GetText(ch.LKEY_COMMANDS_FONTSIZE));
	ch:Println(ch:GetText(ch.LKEY_COMMANDS_OVERHEAL));
end

-- event handling
eventFrame:SetScript("OnEvent", function(self, event, arg1)
	if event then
		if event == "ADDON_LOADED" and arg1 == "ClassicHeals" and not initFinished then
			ch.Init();
		elseif event == "GROUP_ROSTER_UPDATE" or event == "RAID_ROSTER_UPDATE" then
			ch:ResetMemberInfo();
		elseif event == "PLAYER_TARGET_CHANGED" then
			if targetFrame then
				UIFrameFadeOut(targetFrame, 0, 1, 0);
			end
		elseif event == "COMBAT_LOG_EVENT_UNFILTERED" then
			self:OnEvent(event, CombatLogGetCurrentEventInfo());
		elseif event == "PLAYER_ENTERING_WORLD" then
			ch:SetCrfVisible(CLASSICHEALS_OPTIONS[OPT_SHOWCRF]);
		end
	end
end);

-- workaround for changing raid frame interface settings
--_G["InterfaceOptionsFrameOkay"]:HookScript("OnClick", function()
--	ch:ResetMemberInfo();
--end);

function eventFrame:OnEvent(event, ...)
	local timestamp, subevent, _, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags = ...;

	if sourceName == charName and subevent == "SPELL_HEAL" then
		spellId, spellName, spellSchool, amount, overhealing, absorbed, critical = select(12, ...);

		-- party/raid frame
		if inRaid then
			local raidId, subGroupId, playerInSubGroupId = ch:GetAffectedRaidMemberId(destName);
			if raidId and subGroupId and playerInSubGroupId then
				-- compactraidgroup
				if _G["CompactRaidFrameContainer"]:IsVisible() then
					if CompactRaidFrameManager_GetSetting("KeepGroupsTogether") then
						if _G["CompactRaidGroup" .. subGroupId .. "Member" .. playerInSubGroupId] then
							ch:DisplayHeal(raidFrames_Compact[raidId], _G["CompactRaidGroup" .. subGroupId .. "Member" .. playerInSubGroupId], amount, overhealing);
						end
					else
						-- compactraidframe
						local compactRaidFrameId = ch:GetCompactRaidFrameId(destName);
						if compactRaidFrameId and _G["CompactRaidFrame" .. compactRaidFrameId] then
							ch:DisplayHeal(raidFrames_Compact[raidId], _G["CompactRaidFrame" .. compactRaidFrameId], amount, overhealing);
						end
					end
				end
				
				-- pulloutgroup
				local pulloutGroupId = ch:GetRaidPulloutId(subGroupId);
				if pulloutGroupId and _G["RaidPullout" .. pulloutGroupId .. "Button" .. playerInSubGroupId] then
					ch:DisplayHeal(raidFrames_Pullout[raidId], _G["RaidPullout" .. pulloutGroupId .. "Button" .. playerInSubGroupId], amount, overhealing);
				end
			end
		else
			local partyMemberId = ch:GetAffectedPartyMemberId(destName);
			if partyMemberId then
				ch:DisplayHeal(partyFrames[partyMemberId], _G["PartyMemberFrame" .. partyMemberId], amount, overhealing);
			end
		end
			
		-- target frame
		if destName == GetUnitName("target") then
			if not targetFrame then
				targetFrame = ch:CreateFrame("target");
			end

			targetFrame:Show();
			ch:DisplayHeal(targetFrame, _G["TargetFrame"], amount, overhealing);
		end
	end
end

-- main stuff
function ch:SetCrfVisible(value)
	if _G["CompactRaidFrameContainer"] and _G["CompactRaidFrameContainer"]:IsVisible() ~= value then
		if value then
			_G["CompactRaidFrameContainer"]:Show();
		else
			_G["CompactRaidFrameContainer"]:Hide();
		end
	end
end

function ch:DisplayHeal(frame, parentFrame, amount, overhealing)	
	local message;
	local framePositionSet = false;
	
	if overhealing > 0 then
		if CLASSICHEALS_OPTIONS[OPT_SHOWOVERHEAL] then
			ch:SetFramePosition(frame, parentFrame, true, true);
			framePositionSet = true;
			message = "+" .. amount .. " (" .. overhealing .. ")";
		else
			local totalAmount = amount - overhealing;					
			if totalAmount > 0 then
				message = "+" .. totalAmount;
			else
				message = ch:GetText(ch.LKEY_FULL);
			end
		end
	else
		message = "+" .. amount;
	end
	
	if not framePositionSet then
		-- set default position
		ch:SetFramePosition(frame, parentFrame, false, false);
	end

	ch:SetFontSize(frame, critical);
	if critical then
		UIFrameFadeIn(frame, 0.1, 0, 1);
		frame.fadeInfo.finishedFunc = function()
			UIFrameFadeOut(frame, 3, 3, 0);
		end
	else
		UIFrameFadeIn(frame, 0.3, 0, 1);
		frame.fadeInfo.finishedFunc = function()
			UIFrameFadeOut(frame, 1.5, 1.5, 0);
		end
	end
	
	frame:AddMessage(message, 0, 1, 0, nil, 1);
end

function ch:SetFramePosition(frame, parentFrame, hasOverheal, showOverheal)
	if frame:GetName() == "target" then
		frame:SetPoint("CENTER", parentFrame, 41, 2);
	else
		if inRaid then
			if ch:StartsWith(parentFrame:GetName(), "RaidPullout") then
				frame:SetPoint("TOP", parentFrame, 0, 2);
			else
				frame:SetPoint("TOP", parentFrame, 0, -4);
			end
		else
			frame:SetPoint("CENTER", parentFrame, -41, -2);
		end
	end
end

function ch:SetFontSize(frame, critical)
	local size = CLASSICHEALS_OPTIONS[OPT_FONTSIZE];
	
	if critical then
		size = size + 4;
	end

	if inRaid and frame:GetName() ~= "target" then
		size = size - 4;
	end
	
	frame:SetFont("Fonts\\FRIZQT__.TTF", size, "OUTLINE");
end

-- stuff for parties
function ch:GetPartyMembers()
	if not partyMembers then
		partyMembers = {};
		local plist = GetHomePartyInfo();
		
		if plist then
			local n = 1;
			for _, v in pairs(plist) do
				partyMembers[n] = v;
				n = n + 1;
			end
		end
		
		ch:CreatePartyFrames();
	end
	return partyMembers;
end

function ch:CreatePartyFrames()
	for k, _ in pairs(ch:GetPartyMembers()) do
		if not partyFrames[k] then
			partyFrames[k] = ch:CreateFrame("partyFrame" .. k);
		end
	end
end

function ch:GetAffectedPartyMemberId(memberName)
	for playerId, playerName in ipairs(ch:GetPartyMembers()) do
		if playerName == memberName then
			return playerId;
		end
	end

	return nil;
end

-- stuff for raids
function ch:GetRaidMembers()
	if not raidMembers then
		raidMembers = {};
		raidMembers[0] = {};
		local noRaidMembers = GetNumGroupMembers();

		if noRaidMembers then
			for n = 1, noRaidMembers do
				local name, rank, subgroup, level, class, fileName, zone, online, isDead, role, isML = GetRaidRosterInfo(n);
				
				-- table of all raid members by id
				raidMembers[0][n] = name;
				
				-- table of all raid members, mapped by subgroups
				if not raidMembers[subgroup] then
					raidMembers[subgroup] = {};
				end
				local subGroupId = table.getn(raidMembers[subgroup]) + 1;
				raidMembers[subgroup][subGroupId] = name;
				
				n = n + 1;
			end
		end
		
		ch:CreateRaidFrames();
	end
	return raidMembers;
end

function ch:CreateRaidFrames()
	local rm = ch:GetRaidMembers();
	for k, _ in pairs(rm[0]) do
		if not raidFrames_Compact[k] then
			raidFrames_Compact[k] = ch:CreateFrame("raidFrame_Compact" .. k);
		end
		if not raidFrames_Pullout[k] then
			raidFrames_Pullout[k] = ch:CreateFrame("raidFrame_Pullout" .. k);
		end
	end
end

-- returns raidId, subGroupId, playerInSubGroupId
function ch:GetAffectedRaidMemberId(memberName)
	local rm = ch:GetRaidMembers();
	local raidId;
	
	for playerId, playerName in ipairs(rm[0]) do
		if playerName == memberName then
			raidId = playerId;
			break;
		end
	end
	
	for subGroupId, subGroup in pairs(rm) do
		if subGroupId > 0 then
			for playerInSubGroupId, playerName in ipairs(rm[subGroupId]) do
				if playerName == memberName then
					return raidId, subGroupId, playerInSubGroupId;
				end;
			end;
		end
	end

	return nil;
end

function ch:GetCompactRaidFrameId(memberName)
	for n = 1, 40 do
		local compactRaidFrameName = _G["CompactRaidFrame" .. n .. "Name"];
		if compactRaidFrameName and compactRaidFrameName:GetText() == memberName then
			return n;
		end
	end
	
	return nil;
end

function ch:GetRaidPulloutId(subGroupId)
	for n = 1, 8 do
		local raidPulloutName = _G["RaidPullout" .. n .. "Name"];
		if raidPulloutName then
			local pulloutText = raidPulloutName:GetText();
			local textLen = string.len(pulloutText);
			local pulloutGroupId = tonumber(string.sub(pulloutText, textLen, textLen + 1));
			if pulloutGroupId == subGroupId then
				return n;
			end
		end
	end
	
	return nil;
end

-- misc stuff
function ch:Init()
	if not CLASSICHEALS_OPTIONS then
		CLASSICHEALS_OPTIONS = {};
	end
	
	if CLASSICHEALS_OPTIONS[OPT_SHOWCRF] == nil then
		CLASSICHEALS_OPTIONS[OPT_SHOWCRF] = true;
	end
	
	if not CLASSICHEALS_OPTIONS[OPT_FONTSIZE] then
		CLASSICHEALS_OPTIONS[OPT_FONTSIZE] = 16;
	end
	
	if CLASSICHEALS_OPTIONS[OPT_SHOWOVERHEAL] == nil then
		CLASSICHEALS_OPTIONS[OPT_SHOWOVERHEAL] = false;
	end

	initFinished = true;
	ch:Println(ch:GetText(ch.LKEY_ADDONLOADED));
end

function ch:CreateFrame(name)
	local newFrame = CreateFrame("MessageFrame", name, UIParent);
	newFrame:SetWidth(160);
	newFrame:SetHeight(22);
	newFrame:SetFrameStrata("HIGH");
	newFrame:Show();
	return newFrame;
end

function ch:Println(str)
	if str then
		print("[ClassicHeals] " .. str);
	end
end

function ch:StartsWith(str, substr)
   return string.sub(str, 1, string.len(substr)) == substr;
end

function ch:ResetMemberInfo()
	partyMembers = nil;
	raidMembers = nil;
	inRaid = IsInRaid();
end

function ch:BoolToYesNo(value)
	if value then
		return ch:GetText(ch.LKEY_YES);
	end
	return ch:GetText(ch.LKEY_NO);
end